/*
 * Catweasel -- Advanced Floppy Controller
 * Linux device driver
 *
 * Copyright (C) 1998-2009 Michael Krause
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/version.h>

#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/errno.h>
#include <linux/major.h>

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/interrupt.h>

#include <linux/vmalloc.h>
#include <linux/genhd.h>
#include <linux/blkdev.h>

#include <asm/uaccess.h>
#include <asm/io.h>

#define KERNEL_SECTOR_SIZE 512

#if defined(CONFIG_PCI)
#include <linux/pci.h>
#endif

#include "cwfloppy.h"
#if defined(CONFIG_PCI)
#include "mk4_rel2f2.h"
#endif
#include "encoder-common.h"

/* module parameters */
int verbose = 0;
int readonly = 1;
int io[2] = { 0, 0 };
int io_cnt;

MODULE_LICENSE("GPL");
module_param(verbose, int, 0);
//module_param(readonly, int, 0);
module_param_array(io, int, &io_cnt, 0);

static struct cw_controller_struct controllers[2];

static void cwfloppy_request(struct request_queue *q);

/* --------- */

extern struct cw_codec cw_codec_msdos, cw_codec_amiga;

const struct cwfloppy_format cw_formats[] = {
    { "msdosdd",
      512, 1440, 9, 2, 80, 0,
      1, 250, 1, &cw_codec_msdos },
    { "msdoshd",
      512, 2*1440, 18, 2, 80, 0,
      0, 240, 1, &cw_codec_msdos },
    { "amigadd",
      512, 1760, 11, 2, 80, 0,
      1, 250, 0, &cw_codec_amiga },
    { "amigahd",
      512, 2*1760, 22, 2, 80, 0,
      0, 250, 0, &cw_codec_amiga }
};

#define CW_NUM_FORMATS (sizeof(cw_formats) / sizeof(cw_formats[0]))

static struct {
    char verbose, current_format, lastlevel;
    char track, side;
    char count[3];
} errors;

/* -- LOW-LEVEL CATWEASEL FUNCTIONS ------------------------------------- */

static __inline__ void
CWSetCReg (catweasel_contr *c,
	   unsigned char clear,
	   unsigned char set)
{
    c->control_register = (c->control_register & ~clear) | set;
    outb(c->control_register, c->io_sr);
}


void cw_codec_error(int level, const char *fmt, ...)
{
    char buf[50];
    va_list args;

    va_start(args, fmt);
    vsprintf(buf, fmt, args);
    va_end(args);

    if(errors.verbose || level < errors.lastlevel)
	printk(DEVICE_NAME " <%s %d/%d> L%d: %s\n",
	      cw_formats[(int)errors.current_format].name,
	      errors.track,
	      errors.side,
	      level,
	      buf);

    errors.lastlevel = level;
    errors.count[level]++;
}

/* Set current drive select mask */
void catweasel_select(catweasel_contr *c, int dr0, int dr1)
{
    CWSetCReg(c, dr0*c->crm_sel0 | dr1*c->crm_sel1, (!dr0)*c->crm_sel0 | (!dr1)*c->crm_sel1);
}

/* Start/stop the drive's motor */
void catweasel_set_motor(catweasel_drive *d, int on)
{
    if(d->number == 0) {
	CWSetCReg(d->contr, on*d->contr->crm_mot0, (!on)*d->contr->crm_mot0);
    } else {
	CWSetCReg(d->contr, on*d->contr->crm_mot1, (!on)*d->contr->crm_mot1);
    }
}

/* Check if disk in selected drive is write protected. */
int catweasel_write_protected(catweasel_drive *d)
{
    return !(inb(d->contr->io_sr) & 8);
}

static void
catweasel_read_start (catweasel_drive *d,
		      int side,
		      int clock)
{
    int iobase = d->contr->iobase;

    CWSetCReg(d->contr, d->contr->crm_side, (!side) * d->contr->crm_side); /* set disk side */

    if(d->contr->type == CATWEASEL_TYPE_MK1) {
	inb(iobase+1); /* ra reset */
	outb(clock*128, iobase+3);

	inb(iobase+1);
	inb(iobase+0);
	inb(iobase+0);
	outb(0, iobase+3); /* don't store index pulse */

	inb(iobase+1);

	inb(iobase+7); /* start reading */
    } else {
	outb(0, iobase + 0xe4); /* Reset memory pointer */
	switch(clock) {
	case 0: /* 28MHz */
	    outb(128, iobase + 0xec);
	    break;
	case 1: /* 14MHz */
	    outb(0, iobase + 0xec);
	    break;
	}
	inb(iobase + 0xe0);
	inb(iobase + 0xe0);
	outb(0, iobase + 0xec); /* no IRQs, no MFM predecode */
	inb(iobase + 0xe0);
	outb(0, iobase + 0xec); /* don't store index pulse */

	outb(0, iobase + 0xe4); /* Reset memory pointer */
	inb(iobase + 0xf0); /* start reading */
    }
}

static void
catweasel_read_stop (catweasel_drive *d)
{
    int iobase = d->contr->iobase;

    if(d->contr->type == CATWEASEL_TYPE_MK1) {
	outb(0, iobase+1); /* stop reading, don't reset RAM pointer */

	outb(128, iobase+0); /* add data end mark */
	outb(128, iobase+0);
    } else {
	inb(iobase + 0xe4); /* stop reading, don't reset RAM pointer */

	outb(128, iobase + 0xe0); /* add data end mark */
    }
}

static void
catweasel_reset_mem_pointer (catweasel_drive *d)
{
    int iobase = d->contr->iobase;

    if(d->contr->type == CATWEASEL_TYPE_MK1) {
	inb(iobase+1); /* Reset RAM pointer */
	inb(iobase); /* ignore first byte */
    } else {
	outb(0, iobase + 0xe4); /* Reset memory pointer */
    }
}

/* Returns true if the disk in the selected drive has been removed
 * since the last seeking step, even if one has been re-inserted in
 * the meantime.
 *
 * (when this happens, we deny any operations on this drive until it
 * has been closed by all users. Also, we need to do a recalibration
 * when the device is opened.)
 *
 */
static int
catweasel_disk_changed (catweasel_drive *d)
{
    return ! (inb(d->contr->io_sr) & d->contr->srm_dchg);
}

static void
catweasel_write_start (catweasel_drive *d,
		       int side,
		       int clock,
		       int indexed)
{
    int iobase = d->contr->iobase;

    CWSetCReg(d->contr, d->contr->crm_side, (!side) * d->contr->crm_side); /* set disk side */

    /* !! IMPORTANT !! For the latest MK4 firmware, the
       write-enable procedure has changed. If we ever add write
       support to this driver again, take care to check the
       mailing-list archives... !!!!!!!!!!!!!!!!!!!!!!!!!!!! */

    if(d->contr->type == CATWEASEL_TYPE_MK1) {
	inb(iobase+1);
	outb(clock*128, iobase+3);
	inb(iobase);
	outb(128, iobase+3); /* write enable */
	inb(iobase);
	inb(iobase);
	inb(iobase);
	inb(iobase);
	inb(iobase);
	inb(iobase);

	if(!indexed) {
	    outb(0, iobase+5);
	} else {
	    outb(0, iobase+7);
	}
    } else {
	outb(0, iobase + 0xe4); /* Reset memory pointer */
	switch(clock) {
	case 0: /* 28MHz */
	    outb(128, iobase + 0xec);
	    break;
	case 1: /* 14MHz */
	    outb(0, iobase + 0xec);
	    break;
	}
	inb(iobase + 0xe0);
	outb(128, iobase + 0xec); /* write enable */
	inb(iobase + 0xe0);
	inb(iobase + 0xe0);
	inb(iobase + 0xe0);
	inb(iobase + 0xe0);
	inb(iobase + 0xe0);
	inb(iobase + 0xe0);

	if(!indexed) {
	    outb(0, iobase + 0xf4); /* unconditional write START */
	} else {
	    outb(0, iobase + 0xf0); /* indexed write START */
	}
    }
}

static void
catweasel_write_stop (catweasel_drive *d)
{
    int iobase = d->contr->iobase;

    if(d->contr->type == CATWEASEL_TYPE_MK1) {
	inb(iobase+1);
    } else {
	inb(iobase + 0xe4); /* Abort writing */
	inb(iobase + 0xe4); /* Reset memory pointer */
    }
}

/*** READ/WRITE STATE MACHINE ********************************************************/

#define CW_DELAY(function, msecs) \
  PREPARE_DELAYED_WORK(&c->worker, function); \
  queue_delayed_work(c->workqueue, &c->worker, msecs_to_jiffies(msecs));

#define CW_DELAY_JIFFIES(function, jiffies) \
  PREPARE_DELAYED_WORK(&c->worker, function); \
  queue_delayed_work(c->workqueue, &c->worker, jiffies);

/* Track "recalibration": first step inside 3 times, then step outside
 * until we reack track zero. */

static void
cw_process_recalibrate (struct work_struct *work)
{
    struct cw_controller_struct *c = container_of(work, struct cw_controller_struct, worker.work);
    struct request *req = c->req;
    struct cw_drive_struct *d = req->rq_disk->private_data;

    if(d->sm_recalibrate.step == 0 && d->sm_recalibrate.state >= 3) {
	if(!(inb(d->c->c.io_sr) & d->c->c.srm_trk0)) {
	    /* We're at track 0 now. */
	    *d->sm_recalibrate.err = 1;
	    d->sm_recalibrate.state = -1;
	    d->sm_recalibrate.func(work);
	    return;
	}
	
	if(d->sm_recalibrate.state == 3 + 86) {
	    /* We don't seem to be able to get to track 0. */
	    *d->sm_recalibrate.err = 0;
	    d->sm_recalibrate.state = -1;
	    d->sm_recalibrate.func(work);
	    return;
	}
    }

    /* set step direction */
    if(d->sm_recalibrate.state < 3) {
	CWSetCReg(&d->c->c, d->c->c.crm_dir, 0); /* step inside */
    } else {
	CWSetCReg(&d->c->c, 0, d->c->c.crm_dir); /* step outside */
    }

    /* trigger step (requires two writes two the control register with
     * a pause in between) */
    if(d->sm_recalibrate.step == 0) {
	CWSetCReg(&d->c->c, d->c->c.crm_step, 0);
	d->sm_recalibrate.step = 1;
    } else {
	CWSetCReg(&d->c->c, 0, d->c->c.crm_step);
	d->sm_recalibrate.step = 0;
	d->sm_recalibrate.state++;
    }

    CW_DELAY(cw_process_recalibrate, 5);
}

static void
cw_process_recalibrate_0 (struct work_struct *work,
			  void (*func)(struct work_struct *), /* function to call when finished [in no context] */
			  int *err) /* where to put the error code */
{
    struct cw_controller_struct *c = container_of(work, struct cw_controller_struct, worker.work);
    struct request *req = c->req;
    struct cw_drive_struct *d = req->rq_disk->private_data;

    BUG_ON(d->sm_recalibrate.state != -1);

//  printk("cw_process_recalibrate called.\n");

    d->sm_recalibrate.func = func;
    d->sm_recalibrate.err = err;
    d->sm_recalibrate.state = 0;
    d->sm_recalibrate.step = 0;

    cw_process_recalibrate(work);
}

static void
cw_process_seek (struct work_struct *work)
{
    struct cw_controller_struct *c = container_of(work, struct cw_controller_struct, worker.work);
    struct request *req = c->req;
    struct cw_drive_struct *d = req->rq_disk->private_data;

    if(d->sm_seek.recalibrate) {
	if(*d->sm_seek.err == 0) {
	    /* Recalibration failed. */
	    /* ?? Remember this for the future => disable drive or recalibrate every time */
	    d->sm_seek.state = -1;
	    d->sm_seek.func(work);
	    return;
	}
	d->current_track = 0;
	d->sm_seek.recalibrate = 0;
    }

//  printk("seek %d %d\n", d->current_track, d->sm_seek.track);

    if(d->current_track != d->sm_seek.track) {
	int stepdir = (d->sm_seek.track > d->current_track) ? 1 : -1;

	if(stepdir > 0) {
	    CWSetCReg(&d->c->c, d->c->c.crm_dir, 0); /* step inside */
	} else {
	    CWSetCReg(&d->c->c, 0, d->c->c.crm_dir); /* step outside */
	}

	if(d->sm_seek.state == 0) {
	    CWSetCReg(&d->c->c, d->c->c.crm_step, 0);
	    d->sm_seek.state = 1;
	} else {
	    CWSetCReg(&d->c->c, 0, d->c->c.crm_step);
	    d->sm_seek.state = 0;
	    d->current_track += stepdir;
   	}

	CW_DELAY(cw_process_seek, 5);
	return;
    }

    d->sm_seek.state = -1;
    *d->sm_seek.err = 1;
    d->sm_seek.func(work);
}

static void
cw_process_seek_0 (struct work_struct *work,
		   int track, /* seek to this track */
		   int recalibrate, /* include a track0 recalibration? */
		   work_func_t func, /* function to call when finished [in no context] */
		   int *err) /* where to put the error code */
{
    struct cw_controller_struct *c = container_of(work, struct cw_controller_struct, worker.work);
    struct request *req = c->req;
    struct cw_drive_struct *d = req->rq_disk->private_data;

//  printk("cw_process_seek to %d.\n", track);

    BUG_ON(d->sm_seek.state != -1);

    d->sm_seek.track = track;
    d->sm_seek.recalibrate = recalibrate;
    d->sm_seek.func = func;
    d->sm_seek.err = err;
    d->sm_seek.state = 0;

    if(recalibrate) {
	cw_process_recalibrate_0(work, cw_process_seek, err);
    } else {
	cw_process_seek(work);
    }
}

/* Check whether some drive motor is to be switched off. Call this "as
 * often as possible" from the state-machine functions. */
static void
cw_stop_motors (struct cw_controller_struct *c)
{
    int j;

    if(c->c.type != CATWEASEL_TYPE_NONE) {
	for(j = 0; j < 2; j++) {
	    if(c->drives[j].d->type != 0 && c->drives[j].motor_please_stop) {
		catweasel_set_motor(c->drives[j].d, 0);
//		printk(DEVICE_NAME ": cw_stop_motors() stopping motor %08x\n",
//		                   (unsigned int)(c->drives + j));
		c->drives[j].motor_on = 0;
		c->drives[j].motor_please_stop = 0;
	    }
	}
    }
}

static void
motor_off_callback (unsigned long n)
{
    struct cw_drive_struct *d = (struct cw_drive_struct *)n;
    unsigned long flags;

    if(!d->motor_on)
	return;

    spin_lock_irqsave(&d->c->lock, flags);
    if(d->c->busy) {
	/* A request is being processed on the controller belonging to
	 * the specified drive 'd'. Obviously we can't mess around
	 * with the controller's registers right now, so we ask the
	 * request-processing state machine to switch off the driver
	 * motor when it's got some time. */
	d->motor_please_stop = 1;
    } else {
//	printk(DEVICE_NAME ": motor_off_callback() stopping motor %08x\n", (unsigned int)d);
	catweasel_set_motor(d->d, 0);
	d->motor_on = 0;
	d->motor_please_stop = 0;
    }
    spin_unlock_irqrestore(&d->c->lock, flags);
}

static void
cw_motor_off (struct cw_drive_struct *d)
{
    /* Start motor spin down timer. We don't stop the motor now in
       order to avoid superfluous motor start up times for requests
       that follow immediately. */
    del_timer_sync(&d->motor_off_timer);
    d->motor_off_timer.expires = jiffies + 1 * HZ;
    add_timer(&d->motor_off_timer);
}

static void
cw_motor_on (struct cw_drive_struct *d)
{
    del_timer_sync(&d->motor_off_timer);
    d->motor_please_stop = 0;
    if(!d->motor_on) {
	catweasel_set_motor(d->d, 1);
	d->motor_on_jiffies = jiffies;
    }
}

static void
cw_process_rw_track (struct work_struct *work)
{
//    static int state_timeout; /* state variable: timeout counter for write waiting */
    static const int threshoffsets[5] = { 0, -1, +1, -2, +2 };

    int i, mynice;

    struct cw_controller_struct *c = container_of(work, struct cw_controller_struct, worker.work);
    struct request *req = c->req;
    struct cw_drive_struct *d = req->rq_disk->private_data;

  repeat:

//  printk("rw %d / %d\n", d->sm_rw_track.state, d->sm_rw_track.try);

    switch(d->sm_rw_track.state) {
    case 0:
	/* Seek to the desired position. Recalibrate for third or
           later attempts. */
//	printk("00 state_try = %d.\n", d->sm_rw_track.try);
	d->sm_rw_track.state = 1;
//	printk(DEVICE_NAME ": Seeking. Try %d. Recal=%d\n", d->sm_rw_track.try, d->sm_rw_track.try >= 2);
	catweasel_select(&d->c->c, d->d->number == 0, d->d->number == 1);

	cw_motor_on(d);
	/* Start seeking. If device has just been opened
	   (d->justopened == -1), then we force a recalibration to
	   make sure that the disk-change status bit is
	   up-to-date. */
	cw_process_seek_0(work, d->sm_rw_track.track / 2,
			  d->justopened == 1 || d->sm_rw_track.try >= 2,
			  cw_process_rw_track, d->sm_rw_track.err);
	return;

    case 1:
	if(*d->sm_rw_track.err == 0) {
	    /* Seek failed. */
	    printk(DEVICE_NAME ": seek failed.\n");
	    goto ende;
	}

	d->justopened = 0;

//	printk("1a\n");

	/* Seeked to correct position. Wait for motor to spin up
           completely now. */
	if(!d->motor_on) {
	    int t;

	    t = jiffies - d->motor_on_jiffies;
	    if(t >= 500*HZ/1000 || t < 0) {
		d->motor_on = 1;
	    }
	    if(!d->motor_on) {
//		printk("1b\n");
		CW_DELAY_JIFFIES(cw_process_rw_track, 500*HZ/1000 - t);
		return;
	    }
	}

	if(catweasel_disk_changed(d->d)) {
	    printk(DEVICE_NAME ": disk removed / changed.\n");
	    *d->sm_rw_track.err = 0;
	    goto ende;
	}

	if(!d->sm_rw_track.rw) {
	    /* Start reading the track now. */
//	    printk("1c\n");

	    catweasel_read_start(d->d, d->sm_rw_track.track % 2, cw_formats[d->format].clock);
//	    printk(DEVICE_NAME ": Reading track... (try %d)\n", d->sm_rw_track.try);
	    d->sm_rw_track.state = 2;
	    CW_DELAY(cw_process_rw_track, cw_formats[d->format].readtime);
	} else {
	    /* Encode track and start writing. */

	    /* reset catweasel RAM pointer */
	    if(d->c->c.type == CATWEASEL_TYPE_MK1) {
		inb(d->c->c.iobase + 1);
	    } else {
		outb(0, d->c->c.iobase + 0xe4);
	    }
	    /* encode */
	    cw_formats[d->format].codec->encode(d, d->sm_rw_track.track % 2);
	    outb(0xff, d->c->c.io_mem); /* write end mark */
	    /* start writing */
	    catweasel_write_start(d->d, d->sm_rw_track.track % 2,
				  cw_formats[d->format].clock, cw_formats[d->format].indexed);
//	    printk(DEVICE_NAME ": Writing track... (try %d)\n", d->sm_rw_track.try);
	    d->sm_rw_track.state = 4;
	    d->sm_rw_track.state_timeout = 30;
	    CW_DELAY(cw_process_rw_track, 10);
	}
	return;

    case 2:
	/* Track read by Catweasel. Now try to decode. */
	/* This part of the handler runs under keventd (the scheduler
           task queue), so we may safely take some time here. */
	catweasel_read_stop(d->d);
//	printk(DEVICE_NAME ": Track read.\n");
	catweasel_select(&d->c->c, 0, 0);

	errors.lastlevel = 3;
	errors.count[0] = 0;
	errors.count[1] = 0;
	errors.count[2] = 0;
	errors.current_format = d->format;
	errors.track = d->sm_rw_track.track / 2;
	errors.side = d->sm_rw_track.track % 2;

	for(i = 0; (d->sm_rw_track.try >= 1 && i < sizeof(threshoffsets)/sizeof(threshoffsets[0])) || i < 1; i++) {
	    catweasel_reset_mem_pointer(d->d);

	    d->bufferedtrack = -1;
	    /* During decoding of the track, set our nice value to
	       zero. In Linux 2.6.7, this prevented the X pointer from
	       making choppy movements (waitqueue threads had nice -10
	       by default). In Linux 2.6.23, this doesn't help any
	       longer?!? */
	    mynice = task_nice(current);
	    set_user_nice(current, 0);
    	    *d->sm_rw_track.err = cw_formats[d->format].codec->decode
		                     (d, d->sm_rw_track.track % 2, threshoffsets[i]);
	    set_user_nice(current, mynice);

	    if(*d->sm_rw_track.err == 1) {
		/* Successfully decoded. */
		d->bufferedtrack = d->sm_rw_track.track;
		goto ende;
	    }
	}

	/* Track could not be decoded without errors. */
//	printk("no success decoding.\n");
	if(++d->sm_rw_track.try < 5) {
	    /* next try */
	    d->sm_rw_track.state = 0;
//	    printk("00 repeat state_try = %d.\n", d->sm_rw_track.try);
	    goto repeat;
	} else {
	    /* give up */
	    d->defectivetrack = d->sm_rw_track.track;
	    goto ende;
	}
	break;


    case 4:
	/* Triggered Catweasel to start writing. Now check if it is
	 * really doing so. This can take some time in case we're
	 * using 'indexed' writing mode, i.e., writing starts as soon
	 * as the index-hole pulse comes. */

//	printk(DEVICE_NAME ": Waiting for writing to start.\n");
	if(0 != (inb(d->c->c.io_sr) & d->c->c.srm_writ)) {
	    /* Controller is not yet writing. */
	    if(--d->sm_rw_track.state_timeout == 0) {
		/* Time out. Should Not Happen (TM). */
		catweasel_write_stop(d->d);
		printk(DEVICE_NAME ": Catweasel doesn't start writing after 300ms. Retrying...\n");

		if(++d->sm_rw_track.try < 3) {
		    d->sm_rw_track.state = 1;
		    goto repeat;
		} else {
		    *d->sm_rw_track.err = 0;
		    goto ende;
		}
	    }
	} else {
	    d->sm_rw_track.state = 5;
	    d->sm_rw_track.state_timeout = 30;
	}
	CW_DELAY(cw_process_rw_track, 10);
	return;

    case 5:
	/* Wait for writing to finish */
//	printk(DEVICE_NAME ": Waiting for writing to finish (timeout: %d).\n",
//	       d->sm_rw_track.state_timeout * 10);
	if(0 == (inb(d->c->c.io_sr) & d->c->c.srm_writ)) {
	    /* Controller is still writing. */
	    if(--d->sm_rw_track.state_timeout == 0) {
		/* Time out. Should not happen, if the encoder routines are correct. */
		catweasel_write_stop(d->d);
		printk(DEVICE_NAME ": Catweasel doesn't stop writing after 300ms. Retrying...\n");

		if(++d->sm_rw_track.try < 3) {
		    d->sm_rw_track.state = 1;
		    goto repeat;
		} else {
		    *d->sm_rw_track.err = 0;
		    goto ende;
		}
	    }
	} else {
	    catweasel_write_stop(d->d);
	    catweasel_select(&d->c->c, 0, 0);
	    *d->sm_rw_track.err = 1;
	    goto ende;
	}
	CW_DELAY(cw_process_rw_track, 10);
	return;

    default:
	printk(DEVICE_NAME ": Programming error in cw_process_rw_track()\n");
	break;
    }

    goto repeat;

  ende:

    catweasel_select(&d->c->c, 0, 0);
    cw_motor_off(d);
    d->sm_rw_track.state = -1;
    d->sm_rw_track.func(work);
}

static void
cw_process_rw_track_0 (struct work_struct *work,
		       int track, /* track to read (0..159) */
		       int rw, /* 0 = read from disk, 1 = write to disk */
		       work_func_t func, /* function to call when finished [in no context] */
		       int *err) /* where to put the error code */
{
    struct cw_controller_struct *c = container_of(work, struct cw_controller_struct, worker.work);
    struct request *req = c->req;
    struct cw_drive_struct *d = req->rq_disk->private_data;

    BUG_ON(d->sm_rw_track.state != -1);

    if(!rw) {
	if(d->defectivetrack == track) {
	    /* Tried reading this track previously. Didn't work. */
	    *err = 0;
	    func(work);
	    return;
	}
	if(d->bufferedtrack == track) {
	    /* Track is currently in trackbuffer. */
	    *err = 1;
	    func(work);
	    return;
	}
    }

    /* Start the job. */
    d->sm_rw_track.track = track;
    d->sm_rw_track.rw = rw;
    d->sm_rw_track.func = func;
    d->sm_rw_track.err = err;
    d->defectivetrack = -1;
    d->sm_rw_track.state = 0;
    d->sm_rw_track.try = 0;

    cw_process_rw_track(work);
}

static void cw_process_request_done (struct request *req, int result)
{
    struct request_queue *q = req->q;
    struct cw_drive_struct *drive = req->rq_disk->private_data;
    struct cw_controller_struct *controller = drive->c;
    unsigned long flags;
    int i, j;

//    printk("process_request_done.\n");

    spin_lock_irqsave(q->queue_lock, flags);
    end_request(req, result);
    cw_stop_motors(controller);
    controller->busy = 0;

    /* The following sucks. I can't believe I have to do this: */
    for(i = 0; i < 2; i++) {
	if(controller->drives[i].d->type != 0) {
	    for(j = 0; j < CW_NUM_FORMATS; j++) {
		cwfloppy_request(controller->drives[i].gendisks[j].Queue);
	    }
	}
    }

    spin_unlock_irqrestore(q->queue_lock, flags);
}

static void cw_process_request (struct work_struct *work)
{
    static int state = 0;
    int cnt, block, track, sector, rsec, maxsecs;
    char *data;

    struct cw_controller_struct *c = container_of(work, struct cw_controller_struct, worker.work);
    struct request *req = c->req;
    struct cw_drive_struct *d = req->rq_disk->private_data;

    static int cw_result;

//    printk("process_request\n");

    maxsecs = cw_formats[d->format].max_secs_per_track;
    block = req->sector;
    cnt = req->current_nr_sectors;
    data = req->buffer;

    /* 'cnt' should be positive */
    if(cnt <= 0) {
	printk(DEVICE_NAME ": Request size <= 0.\n");
	goto error;
    }

    track = block / maxsecs;
    sector = block % maxsecs;
	
    rsec = maxsecs - sector; /* remaining sectors on this track */
    if(rsec > cnt)
	rsec = cnt;

    if(block + rsec > cw_formats[d->format].numblocks) {
	printk(DEVICE_NAME ": Access beyond end of device.\n");
	goto error;
    }

    if(rq_data_dir(req) == READ) {
	/* READING */
//	printk(DEVICE_NAME ": READ %d %d\n", block, cnt);
	switch(state) {
	case 0:
	    state = 1;
	    cw_process_rw_track_0(work, track, 0, cw_process_request, &cw_result);
	    return;

	case 1:
	    if(!cw_result) {
		goto error;
	    }

	    memcpy(data, d->trackbuffer + sector*512, rsec*512);

	    req->sector += rsec;
	    req->current_nr_sectors -= rsec;
	    req->buffer += rsec << 9;

	    if(req->current_nr_sectors == 0) {
		/* The request has been fulfilled successfully. */
		state = 0;
		cw_process_request_done(req, 1);
	    } else {
		/* The request includes data from other tracks. Continue. */
		state = 0;
		cw_process_request(work);
	    }
	    return;

	default:
	    printk(DEVICE_NAME ": Programming error in cw_process_request() READ\n");
	    break;
	}
    } else {
	/* WRITING */
	printk(DEVICE_NAME ": WRITE %d %d\n", block, cnt);

	switch(state) {
	case 0: /* Start of state machine: start processing this track. */
	    state = 1;
	    if(!(d->bufferedtrack == track || (sector == 0 && cnt == maxsecs))) {
		/* Read track first, because we are going to change part
		 * of it only. */
//		printk(DEVICE_NAME ": write - read track first!\n");
#if 0
		cw_process_rw_track_0(req, track, 0, (void(*)(struct request *))cw_process_request, &cw_result);
		return;
#else
/*
  nur so fr den anfang, damit ich eine ganze disk beschreiben kann,
  ohne da der track vorher gelesen wird...
*/
		d->bufferedtrack = track;
#endif
	    } else {
		d->bufferedtrack = -1;
	    }

	    cw_result = 1;
	    /* fall through */

	case 1: /* Returns from initial track reading. */
	    if(!cw_result) {
		printk(DEVICE_NAME ": write - couldn't do initial read, schade.\n");
		goto error;
	    }
	    /* Copy data into trackbuffer, write track. */
	    memcpy(d->trackbuffer + sector*512, data, rsec*512);
	    state = 2;
//	    printk(DEVICE_NAME ": write - starting write\n");
	    cw_process_rw_track_0(work, track, 1, cw_process_request, &cw_result);
	    return;

	case 2: /* Returns from track writing. */
	    if(!cw_result) {
		goto error;
	    }

//	    printk(DEVICE_NAME ": write - returned from track writing.\n");

	    req->sector += rsec;
	    req->current_nr_sectors -= rsec;
	    req->buffer += rsec << 9;

	    d->bufferedtrack = track;

	    if(req->current_nr_sectors == 0) {
		/* The request has been fulfilled successfully. */
		state = 0;
		cw_process_request_done(req, 1);
	    } else {
		/* The request includes data from other tracks. Continue. */
		state = 0;
		cw_process_request(work);
	    }
	    return;

	default:
	    printk(DEVICE_NAME ": Programming error in cw_process_request() WRITE\n");
	    break;
	}
    }


  error:
    state = 0;
    cw_process_request_done(req, 0);
}

static void cwfloppy_request(struct request_queue *q)
{
    struct cw_drive_struct *d = q->queuedata;
    struct cw_controller_struct *c = d->c;

    if(c->busy) {
	/* Already busy processing requests for this controller. */
	return;
    }

    while((c->req = elv_next_request(q)) != NULL) {
	if(!blk_fs_request(c->req)) {
	    end_request(c->req, 0);
	    continue;
	}

	switch(rq_data_dir(c->req)) {
	case READ:
	case WRITE:
	    c->busy = 1;
	    CW_DELAY_JIFFIES(cw_process_request, 0);
	    return;

	default:
	    printk(KERN_WARNING "cwfloppy_request: unknown command.\n");
	    end_request(c->req, 0);
	    continue;
	}

	end_request(c->req, 1);
    }
}

// up to Linux 2.6.2x this was:
//static int cwfloppy_open(struct inode *inode, struct file *filp)
// now:        int (*open) (struct block_device *, fmode_t);
static int cwfloppy_open (struct block_device *bdev, fmode_t mode)
{
    int dev, contr, drive, format;
    struct cw_drive_struct *d;
    const struct cwfloppy_format *f;

    dev = MINOR(bdev->bd_dev);
    contr = dev / (2 * CW_MAX_FORMATS);
    drive = (dev - (2 * CW_MAX_FORMATS * contr)) / CW_MAX_FORMATS;
    format = (dev - (CW_MAX_FORMATS * (2 * contr + drive))) - 1;

    if(format < 0 || format >= CW_NUM_FORMATS)
	return -ENODEV;

    d = &controllers[contr].drives[drive];
    f = &cw_formats[format];

    /* disallow simultaneous access to the same drive with different
     * device numbers */
    if(d->in_use && d->format != format) {
	return -EBUSY;
    }

    /* check if disk is in drive. */
#if 0
    return -ENXIO;
#endif

    /* if read-write access is requested, check if disk is
     * write-protected. */
    if(mode & FMODE_WRITE) {
	if(readonly || !f->codec->encode || f->readonly) {
	    return -EROFS;
	}
    }

    f = &cw_formats[format];
    d->in_use++;

    if(d->in_use == 1) {
	/* This device is opened for the first time: */
	d->format = format;
	d->trackbuffer = kmalloc(f->blksize * f->max_secs_per_track, GFP_KERNEL);
	if(!d->trackbuffer) {
	    printk(DEVICE_NAME ": No memory for trackbuffer (minor %d)!\n", dev);
	    d->in_use--;
	    return -ENOMEM;
	}
	d->bufferedtrack = -1;
	d->defectivetrack = -1;
	d->justopened = 1;
    }

#if 0
    /* I don't know if I should call this "elegant" or "a stupid
       hack".  Anyway, it makes lseek() work (UAE needs this, for
       example).  Implementing fops->llseek() instead is overkill IMHO */
    if(filp && filp->f_dentry && filp->f_dentry->d_inode)
	filp->f_dentry->d_inode->i_size = f->numblocks * f->blksize;
#endif

#if 0
    if(filp && filp->f_mode)
	check_disk_change(inode->i_rdev);
#endif

    return 0;
}

static int cwfloppy_release (struct gendisk *gd, fmode_t f)
{
    struct cw_drive_struct *d = gd->private_data;
    
    d->in_use--;
    if(d->in_use == 0) {
	kfree(d->trackbuffer);
    }

    return 0;
}

static struct block_device_operations cwfloppy_ops = {
    .owner      = THIS_MODULE,
//    .ioctl	     = sbd_ioctl
    .open	= cwfloppy_open,
    .release	= cwfloppy_release,
};

/**************************************************************************
 *
 * Stuff below is only used during module initialization / exiting
 *
 **************************************************************************/

static DECLARE_WAIT_QUEUE_HEAD(ms_wait);

static struct timer_list ms_timer;
static int ms_busy;

static void
ms_sleeping_timeout (unsigned long x)
{
    ms_busy = 0;
    wake_up(&ms_wait);
}

static void
ms_sleeping_delay (int ms)
{
    if(ms <= 0)
	return;

    ms_busy = 1;
    ms_timer.function = ms_sleeping_timeout;
    ms_timer.expires = jiffies + msecs_to_jiffies(ms);
    add_timer(&ms_timer);

    while(ms_busy)
	sleep_on(&ms_wait);
}

static void
CWTriggerStepSleepy (catweasel_contr *c)
{
    CWSetCReg(c, c->crm_step, 0);
    ms_sleeping_delay(5);
    CWSetCReg(c, 0, c->crm_step);
    ms_sleeping_delay(5);
}

/* Initialize a Catweasel controller; c->iobase and c->msdelay must have
   been initialized -- msdelay might be used */
static void
catweasel_init_controller (catweasel_contr *c)
{
    int i, j;

    if(!c->iobase)
	return;

    switch(c->type) {
    case CATWEASEL_TYPE_MK1:
	c->crm_sel0 = 1 << 5;
	c->crm_sel1 = 1 << 4;
	c->crm_mot0 = 1 << 3;
	c->crm_mot1 = 1 << 7;
	c->crm_dir  = 1 << 1;
	c->crm_step = 1 << 0;
	c->crm_side = 1 << 2;
	c->srm_trk0 = 1 << 4;
	c->srm_dchg = 1 << 5;
	c->srm_writ = 1 << 1;
	c->io_sr    = c->iobase + 2;
	c->io_mem   = c->iobase;
	break;
    case CATWEASEL_TYPE_MK3:
    case CATWEASEL_TYPE_MK4:
	c->crm_sel0 = 1 << 2;
	c->crm_sel1 = 1 << 3;
	c->crm_mot0 = 1 << 1;
	c->crm_mot1 = 1 << 5;
	c->crm_dir  = 1 << 4;
	c->crm_step = 1 << 7;
	c->crm_side = 1 << 6;
	c->srm_trk0 = 1 << 2;
	c->srm_dchg = 1 << 5;
	c->srm_writ = 1 << 6;
	c->io_sr    = c->iobase + 0xe8;
	c->io_mem   = c->iobase + 0xe0;
	break;
    default:
	return;	
    }

    c->control_register = 255;

    /* select all drives, step inside */
    CWSetCReg(c, c->crm_dir | c->crm_sel0 | c->crm_sel1, 0);
    for(i=0;i<3;i++)
	CWTriggerStepSleepy(c);

    for(i=0;i<2;i++) {
	c->drives[i].number = i;
	c->drives[i].contr = c;
	c->drives[i].diskindrive = 0;

	/* select only the respective drive, step to track 0 */
	if(i == 0) {
	    CWSetCReg(c, c->crm_sel0, c->crm_dir | c->crm_sel1);
	} else {
	    CWSetCReg(c, c->crm_sel1, c->crm_dir | c->crm_sel0);
	}
	
	for(j = 0; j < 86 && (inb(c->io_sr) & c->srm_trk0); j++)
	    CWTriggerStepSleepy(c);
	
	if(j < 86) {
	    c->drives[i].type = 1;
	} else {
	    c->drives[i].type = 0;
	}
    }
    
    CWSetCReg(c, 0, c->crm_sel0 | c->crm_sel1); /* deselect all drives */
}

/* Reset the controller */
static void
catweasel_free_controller (catweasel_contr *c)
{
    if(!c->iobase)
	return;

    /* all motors off, deselect all drives */
    CWSetCReg(c, 0, c->crm_mot0 | c->crm_mot1 | c->crm_sel0 | c->crm_sel1);
}

static void
init_controller_drive (struct cw_drive_struct *d)
{
    init_timer(&d->motor_off_timer);
    d->motor_off_timer.data = (unsigned long)d;
    d->motor_off_timer.function = motor_off_callback;

    d->motor_on = 0;
    d->current_track = 0;
    d->trigger_step_sm = 0;
    d->motor_please_stop = 0;

    d->sm_rw_track.state = -1;
    d->sm_seek.state = -1;
    d->sm_recalibrate.state = -1;

    if(d->d->type)
	printk(DEVICE_NAME ": Detected drive %d @0x%x\n",
	       d->d->number, d->c->c.iobase);
}

static int
init_controller (struct cw_controller_struct *c)
{
    memset(&c->drives, 0, sizeof(c->drives));
    
    catweasel_init_controller(&c->c);

    c->drives[0].d = &c->c.drives[0];
    c->drives[0].c = c;
    init_controller_drive(&c->drives[0]);

    c->drives[1].d = &c->c.drives[1];
    c->drives[1].c = c;
    init_controller_drive(&c->drives[1]);

    spin_lock_init(&c->lock);
    c->busy = 0;

    c->workqueue = create_singlethread_workqueue("cwfloppy");
    INIT_DELAYED_WORK(&c->worker, cw_process_rw_track);

    return 0;
}

static void
free_controller (struct cw_controller_struct *c)
{
    if(c->c.type == CATWEASEL_TYPE_NONE)
	return;

    if(!c->c.iobase)
	return;

    destroy_workqueue(c->workqueue);

    del_timer_sync(&c->drives[0].motor_off_timer);
    del_timer_sync(&c->drives[1].motor_off_timer);
    catweasel_free_controller(&c->c);

    switch(c->c.type) {
    case CATWEASEL_TYPE_MK1:
	release_region(c->c.iobase, 8);
	break;
#if defined(CONFIG_PCI)
    case CATWEASEL_TYPE_MK3:
    case CATWEASEL_TYPE_MK4:
	release_region(c->c.iobase, 256);
	pci_disable_device(c->c.pcidev);
	break;
#endif
    }
}

static int
find_free_controller_struct (void)
{
    int i;

    for(i = 0; i < sizeof(controllers)/sizeof(controllers[0]); i++) {
	if(controllers[i].c.type == CATWEASEL_TYPE_NONE) {
	    return i;
	}
    }

    return -1;
}

static int
add_isa_controller (int iobase)
{
    int err;
    int cn;
    struct cw_controller_struct *c;

    if(!iobase)
	return 0;

    if((cn = find_free_controller_struct()) < 0) {
	printk(KERN_NOTICE DEVICE_NAME ": More than two Catweasels are not supported.\n");
	return -ENODEV;
    }

    c = controllers + cn;
    c->c.type = CATWEASEL_TYPE_MK1;
    c->c.iobase = iobase;

    if(!request_region(c->c.iobase, 8, DEVICE_NAME)) {
	printk(KERN_ERR DEVICE_NAME ": io ports 0x%04x-0x%04x in use\n", c->c.iobase, c->c.iobase+7);
	return -EBUSY;
    }

    printk(KERN_NOTICE DEVICE_NAME ": Controller #%d is an ISA Catweasel @0x%04x\n", cn, iobase);

    if((err = init_controller(c))) {
	release_region(c->c.iobase, 8);
	return err;
    }

    return 0;
}

#if defined(CONFIG_PCI)

static int
cwfloppy_probe_mk3(struct pci_dev *pcidev, const struct pci_device_id *pciid)
{
    int cn;
    struct cw_controller_struct *c;
    int err;

    if((cn = find_free_controller_struct()) < 0) {
	printk(KERN_NOTICE DEVICE_NAME ": More than two Catweasels are not supported.\n");
	return -ENODEV;
    }

    if((err=pci_enable_device(pcidev)))
	return err;

    c = controllers + cn;
    c->c.pcidev = pcidev;
    c->c.type = CATWEASEL_TYPE_MK3;
    c->c.iobase = pci_resource_start(pcidev, 0);

    if(!request_region(c->c.iobase, 256, DEVICE_NAME)) {
	printk(KERN_ERR DEVICE_NAME ": io ports 0x%04x-0x%04x in use\n", c->c.iobase, c->c.iobase+255);
	return -EBUSY;
    }

    printk(KERN_NOTICE DEVICE_NAME ": Controller #%d is a PCI Catweasel MK3 @0x%04x\n", cn, c->c.iobase);

    /* Magic PCI bridge initialization sequence provided by Jens */
    outb(0xf1, c->c.iobase + 0x0);
    outb(0x00, c->c.iobase + 0x1);
    outb(0x00, c->c.iobase + 0x2);
    outb(0x00, c->c.iobase + 0x4);
    outb(0x00, c->c.iobase + 0x5);
    outb(0x00, c->c.iobase + 0x29);
    outb(0x00, c->c.iobase + 0x2b);

    if((err = init_controller(c))) {
	free_controller(c);
	return err;
    }

    return 0;
}

static int
cwfloppy_probe_mk4(struct pci_dev *pcidev, const struct pci_device_id *pciid)
{
    int cn;
    struct cw_controller_struct *c;
    int err;
    unsigned int timeout;
    unsigned char mem[] = { 1,2,4,8,16,32,64,128,255,44,65,3 };

    if((cn = find_free_controller_struct()) < 0) {
	printk(KERN_NOTICE DEVICE_NAME ": More than two Catweasels are not supported.\n");
	return -ENODEV;
    }

    if((err=pci_enable_device(pcidev)))
	return err;

    c = controllers + cn;
    c->c.pcidev = pcidev;
    c->c.type = CATWEASEL_TYPE_MK4;
    c->c.iobase = pci_resource_start(pcidev, 0);

    if(!request_region(c->c.iobase, 256, DEVICE_NAME)) {
	printk(KERN_ERR DEVICE_NAME ": io ports 0x%04x-0x%04x in use\n", c->c.iobase, c->c.iobase+255);
	return -EBUSY;
    }

    printk(KERN_NOTICE DEVICE_NAME ": Controller #%d is a PCI Catweasel MK4 @0x%04x\n", cn, c->c.iobase);

    /* Initialize PCI bridge */
    outb(0xf1, c->c.iobase + 0x0);
    outb(0x00, c->c.iobase + 0x1);
    outb(0xe3, c->c.iobase + CW_DATA_DIRECTION);
    outb(CW_BANK_FLOPPY, c->c.iobase + CW_SELECT_BANK);
    outb(0x00, c->c.iobase + 0x4);
    outb(0x00, c->c.iobase + 0x5);
    outb(0x00, c->c.iobase + 0x29);
    outb(0x00, c->c.iobase + 0x2b);

    /* reset FPGA */
    /*outb(0xe3, c->c.iobase + 0x2);*/
    outb(CW_BANK_RESETFPGA, c->c.iobase + CW_SELECT_BANK);
    udelay(1000);
    outb(CW_BANK_FLOPPY, c->c.iobase + CW_SELECT_BANK);

    /* upload firmware */
    for(err=0; err!=sizeof(cw_firmware); ++err) {
	/* set some direction */
	outb( (cw_firmware[err] & 1)?(CW_BANK_FLOPPY+2):CW_BANK_FLOPPY, c->c.iobase + CW_SELECT_BANK);
	/* wait for FPGA */
	timeout=0;
	while(!(inb(c->c.iobase+7)&8)) {
	    udelay(1);
	    if(++timeout >= 1000000) {
		printk(KERN_ERR DEVICE_NAME ": timeout while writing firmware byte %i\n", err);
		// FIXME: is free_controller(c) okay here?
 		free_controller(c);
		return -EBUSY;
	    }
	}
	/* write byte */
	outb(cw_firmware[err], c->c.iobase + CW_FLOPPY_JOYDAT);
    }

    /* now wait until FPGA really comes alive */
    timeout=0;
    while(inb(c->c.iobase + CW_FLOPPY_CONTROL) == 10) udelay(1), timeout++;

    /* check if memory access is working */
    outb(0, c->c.iobase + CW_FLOPPY_RESETPOINTER);
    for(err=0; err!=sizeof(mem); ++err)
	outb(mem[err], c->c.iobase + CW_FLOPPY_MEMORY);
    outb(0, c->c.iobase + CW_FLOPPY_RESETPOINTER);
    for(err=0; err!=sizeof(mem); ++err) {
	if((timeout=inb(c->c.iobase + CW_FLOPPY_MEMORY)) != mem[err]) {
	    printk(KERN_ERR DEVICE_NAME ": memory test failed. Byte %i should be %02x but is %02x\n", err, mem[err], timeout);
	    // FIXME: is free_controller(c) okay here?
	    free_controller(c);
	    return -EBUSY;
	}
    }

    outb(CW_BANK_FLOPPY, c->c.iobase + CW_SELECT_BANK);

    if((err = init_controller(c))) {
	free_controller(c);
	return err;
    }

    return 0;
}

static struct pci_device_id id_table_mk3[] = {
    { CW_MK4_VENDOR, CW_MK4_DEVICE, 0x1212, 0x0002, },
    { 0, }
};
MODULE_DEVICE_TABLE(pci, id_table_mk3);
static struct pci_driver cwfloppy_pci_mk3 = {
    name: "cwfloppy_mk3",
    id_table: id_table_mk3,
    probe: cwfloppy_probe_mk3
};

static struct pci_device_id id_table_mk4[] = {
    /* The MK4 PCI bridge has a bug where the reported subvendor and
     * subdevice IDs may randomly change between various values. */
    { CW_MK4_VENDOR, CW_MK4_DEVICE, 0x5213, 0x0002, },
    { CW_MK4_VENDOR, CW_MK4_DEVICE, 0x5213, 0x0003, },
    { CW_MK4_VENDOR, CW_MK4_DEVICE, 0x5200, 0x0002, },
    { CW_MK4_VENDOR, CW_MK4_DEVICE, 0x5200, 0x0003, },
    { 0, }
};
MODULE_DEVICE_TABLE(pci, id_table_mk4);
static struct pci_driver cwfloppy_pci_mk4 = {
  .name = "cwfloppy_mk4",
  .id_table = id_table_mk4,
  .probe = cwfloppy_probe_mk4,
  /*.remove = cw_remove,*/
};

#endif /* CONFIG_PCI */

static int
cwfloppy_init_gendisk (struct cw_drive_struct *drive, int minorcount)
{
    static char drivecount = '0';
    int i;

    if(drive->d->type == 0) {
	return 1;
    }

    minorcount++;

    for(i = 0; i < CW_NUM_FORMATS; i++) {
	struct cw_drive_gendisk *gd = drive->gendisks + i;

	gd->gd = alloc_disk(1);
	if(!gd->gd) {
	    return 0;
	}

	gd->Queue = blk_init_queue(cwfloppy_request, &drive->c->lock);
	if(gd->Queue == NULL)
	    return 0;
	gd->Queue->queuedata = drive;
	blk_queue_hardsect_size(gd->Queue, KERNEL_SECTOR_SIZE);

	gd->gd->major = CWFLOPPY_MAJOR;
	gd->gd->first_minor = gd->minor = minorcount++;
	gd->gd->fops = &cwfloppy_ops;
	gd->gd->private_data = drive;
	strcpy(gd->gd->disk_name, "cwX");
	gd->gd->disk_name[2] = drivecount;
	strcat(gd->gd->disk_name, cw_formats[i].name);
	set_capacity(gd->gd, cw_formats[i].numblocks);

	/* In Linux 2.6.7, it looks like we really need a separate
	 * request queue for each gendisk. */
	gd->gd->queue = drive->gendisks[i].Queue;

	add_disk(gd->gd);
    }

    drivecount++;

    return 1;
}

static void
cwfloppy_exit_gendisk (struct cw_drive_struct *drive)
{
    int i;

    if(drive->d->type == 0)
	return;

    for(i = 0; i < CW_NUM_FORMATS; i++) {
	struct cw_drive_gendisk *gd = drive->gendisks + i;

	if(gd->gd != NULL) {
	    del_gendisk(gd->gd);
	    put_disk(gd->gd);
	    blk_cleanup_queue(gd->Queue);
	}
    }
}

static int
cwfloppy_init (void)
{
    int err;

    init_timer(&ms_timer);
    controllers[0].c.type = CATWEASEL_TYPE_NONE;
    controllers[1].c.type = CATWEASEL_TYPE_NONE;

#if defined(CONFIG_PCI)
    /* First try to find a PCI Catweasel */
    err = pci_register_driver(&cwfloppy_pci_mk3);
    if(err && err != -ENODEV)
	return err;
    err = pci_register_driver(&cwfloppy_pci_mk4);
    if(err && err != -ENODEV)
	return err;
    if(controllers[0].c.type == CATWEASEL_TYPE_NONE) {
	printk(KERN_NOTICE DEVICE_NAME ": No PCI Catweasels found.\n");
    }
#endif /* CONFIG_PCI */

    /* Now check if the user has provided an IO base for ISA
       Catweasels */
    err = add_isa_controller(io[0]);
    if(err)
	goto err_isa_failed;
    err = add_isa_controller(io[1]);
    if(err)
	goto err_isa_failed;

    if(controllers[0].c.type == CATWEASEL_TYPE_NONE) {
	printk(KERN_NOTICE DEVICE_NAME ": No ISA Catweasels found. Use \"io=0xNNN\" with insmod\n");
	err = -ENODEV;
	goto err_isa_failed;
    }

    err = register_blkdev(CWFLOPPY_MAJOR, "cwfloppy");
    if(err) {
	printk(KERN_WARNING "cwfloppy: unable to get major number\n");
	goto err_isa_failed;
    }

    codec_msdos_init();

    if((errors.verbose = verbose))
	printk(DEVICE_NAME ": Verbose error reporting enabled\n");
    if(readonly)
	printk(DEVICE_NAME ": Writing disabled for all drives.\n");

    err = -ENOMEM;
    if(!cwfloppy_init_gendisk(controllers[0].drives + 0, 0 * CW_MAX_FORMATS))
	goto err_unregister;
    if(!cwfloppy_init_gendisk(controllers[0].drives + 1, 1 * CW_MAX_FORMATS))
	goto err_unregister;
    if(controllers[1].c.type != CATWEASEL_TYPE_NONE) {
	if(!cwfloppy_init_gendisk(controllers[1].drives + 0, 2 * CW_MAX_FORMATS))
	    goto err_unregister;
	if(!cwfloppy_init_gendisk(controllers[1].drives + 1, 3 * CW_MAX_FORMATS))
	    goto err_unregister;
    }
    return 0;

  err_unregister:
    cwfloppy_exit_gendisk(controllers[0].drives + 0);
    cwfloppy_exit_gendisk(controllers[0].drives + 1);
    if(controllers[1].c.type != CATWEASEL_TYPE_NONE) {
	cwfloppy_exit_gendisk(controllers[1].drives + 0);
	cwfloppy_exit_gendisk(controllers[1].drives + 1);
    }

    unregister_blkdev(CWFLOPPY_MAJOR, "cwfloppy");

  err_isa_failed:
    free_controller(&controllers[1]);
    free_controller(&controllers[0]);
#if defined(CONFIG_PCI)
    pci_unregister_driver(&cwfloppy_pci_mk4);
    pci_unregister_driver(&cwfloppy_pci_mk3);
#endif

    return err;
}

static void
cwfloppy_exit (void)
{
    cwfloppy_exit_gendisk(controllers[0].drives + 0);
    cwfloppy_exit_gendisk(controllers[0].drives + 1);
    if(controllers[1].c.type != CATWEASEL_TYPE_NONE) {
	cwfloppy_exit_gendisk(controllers[1].drives + 0);
	cwfloppy_exit_gendisk(controllers[1].drives + 1);
    }

    free_controller(&controllers[1]);
    free_controller(&controllers[0]);
#if defined(CONFIG_PCI)
    pci_unregister_driver(&cwfloppy_pci_mk4);
    pci_unregister_driver(&cwfloppy_pci_mk3);
#endif

    unregister_blkdev(CWFLOPPY_MAJOR, "cwfloppy");
//    blk_cleanup_queue(Queue);
}

module_init(cwfloppy_init);
module_exit(cwfloppy_exit);
